/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 *  STRINGS.HPP -- STRING class declarations           V2.0  *
 *                                                           *
 *  Operations:                                              *
 *  - Constructs class objects with pointer to character     *
 *    space allocated from the heap and pointer to current   *
 *    position in character array.                           *
 *  - Provides a collection of operators, methods, and       *
 *    friend functions to implement abstract string type,    *
 *    including indexing.                                    *
 *  - Derived class BCstring provides bounds checking for    *
 *    indexing operator.
 *  Assumptions:                                             *
 *  - string.hpp header file will be included with any C++   *
 *    source file wishing to use string services.            *
 *                                                           *
 *  History:                                                 *
 *  - V1.0 - 26 Jan 92                                       *
 *  - V1.1 -  7 Jun 92: removed get_token, unget_token,      *
 *                      added message macro                  *
 *  - V2.0 - 21 Jun 92: added derived class BSstring         *
 *                                                           *
 *  Language:       Borland C++                              *
 *  System:         Gateway 386/25, MS-DOS 5.0               *
 *  Application:    CSE 687                                  *
 *  Author:         Jim Fawcett, CST 2-127, (315)443-4454    *
 *                                                           *
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

#ifndef STRINGS_HPP
#define STRINGS_HPP
#include <string.h>    // ANSI:  strlen(), strcpy(), strcat()
#include <stdlib.h>    // ANSI:  exit();
#include <fstream.h>   // BCC:   includes iostream.h

struct trace {

  char *str;
  trace(char* s) { str = s; cout << "entering " << s << "\n"; }
  ~trace(void)   { cout << "leaving  " << str << "\n"; }

};

class string {

  protected:
    int   slen;      // # of chars in string without null term.
    char* shead;     // pointer to beginning of string
    char* sptr;      // pointer to current position

  public:
    // constructors
    string(void);    // string str;
		     // needed to define an array of objects
    string(const string &s);   // string str1 = str2;
    string(char *s);           // string str(sptr);
    string(int len);           // string str(5);
															 // blank string w/ 5 chars
    // destructor
    virtual ~string(void);
/**/
    // operator methods
    virtual string& operator= (const string &s);    // str2 = str1;
    virtual string& operator+=(const string &s);    // str2 += str1;

    // operator friends
    friend string   operator+ (const string &s1, const string &s2);
			      // concatenation: str3 = str1 + str2
    friend istream& operator>>(istream&, string&);
			      // input from istream
    friend ostream& operator<<(ostream&, string&);
			      // output to istream
    friend      int operator==(const string &str1, const string &str2)
			{ return strcmp(str1.shead, str2.shead) == 0; }
    friend      int operator!=(const string &str1, const string &str2)
			{ return strcmp(str1.shead, str2.shead) != 0; }
    friend      int operator<=(const string &str1, const string &str2)
			{ return strcmp(str1.shead, str2.shead) <= 0; }
    friend      int operator< (const string &str1, const string &str2)
			{ return strcmp(str1.shead, str2.shead) <  0; }
    friend      int operator>=(const string &str1, const string &str2)
			{ return strcmp(str1.shead, str2.shead) >= 0; }
    friend      int operator> (const string &str1, const string &str2)
			{ return strcmp(str1.shead, str2.shead) >  0; }

    // other methods
    virtual char& operator[] (int n) { return *(sptr=shead+n); }
			// subscriptor: x = str[3] = y;
    virtual string overlay(const string &str, int n);
			// overlay str on self starting at index n
    virtual string insert(const string &str, int n);
			// insert str in self starting at index n

    // other friends
    friend int length(const string &s) { return s.slen; }
			// returns string length
    friend char *match(string &str1, const string &str2);
			// find embedded matching substring
    friend int is_in(char ch, const string &set)
			  { return (strchr(set.shead,ch) != NULL); }
			// find embedded matching character
    friend int  index(const string &str)
			  { return int(str.sptr-str.shead); }
			// return current index
    friend int select(string &str, int i=1);
			// return char at current index + i
    friend string head(string &str);
			// return new string equal to head of str
    friend string tail(string &str);
			// return new string equal to tail of str
    friend char*  skipwhite(string &str);
			// iterate down string until get nonwhitespace
    friend string eatwhite(string &str);
			// return str without leading whitespace
    #ifdef VOCAL
      #define message(s) cout << (s)
    #else
      #define message(s)
    #endif
};
/**/

class BCstring : public string {  // strings with bounds checking
				  // simply adds new operator[]
  public:

  // constructors passing parameters to and calling
  // base constructors
    BCstring(void) { };
    BCstring(char *s)     : string(s) { };
    BCstring(string &s)   : string(s) { };
    BCstring(BCstring &s) : string(s) { };
  // destructor calling base destructor
    ~BCstring(void) { };

  // virtual bounds checking indexing function
    char& operator[](int n) {
      if(n<0 || slen<n) {
	cerr << "index [" << n << "] out of range\n";
	exit(1);
      }
      return *(sptr=shead+n);
    }
  // Note:
  // select function needs bounds checking but can not be
  // redefined since it is a friend function and therefore
  // not virtual.  It also can not be overloaded by a new
  // function select with the same signature, e.g. that
  // would be a multiple definition.
};

#endif

